/*
 * Copyright 2010 Cyril Crassin.
 * v1.5
 *
 * This sample shows how dynamic template parameters can be emulated.
 * Cf. http://blog.icare3d.org
*/
#ifndef DYNAMICKERNELTEMPLATES_H
#define DYNAMICKERNELTEMPLATES_H

#define USE_BOOLEAN_SELECTION 0 //Faster to compile but only for boolean parameters

//////////////////////
////Core mechanism////
//////////////////////

//ValueSelector prototype.
//recursively instanciate all possible values and select the right one at runtime
template<typename TList, int NumArguments, class Func, int MaxArgValue, int CurVal>
struct ValueSelector {
	//Value selection function.
	inline static void select(Func *functor, std::vector<int> &templateParamList, int val);
};

//Partial specialization for the last recursion
template<typename TList, int NumArguments, class Func, int MaxArgValue>
struct ValueSelector<TList, NumArguments, Func, MaxArgValue, -1> {
	//Do nothing
	inline static void select(Func *functor, std::vector<int> &templateParamList, int val){
	}
};

//This struct is used to transform a list of parameters into template arguments
//The idea is to build a typelist containing the arguments
//and to pass this typelist to a user defined functor.
template<	typename TList, int NumArguments, class Func, 
			int MaxArgValue> //New maximum value of integer template parameters.
struct KernelCaller {
	//Convenience function used by the user
	//Takes a variable argument list, transforms it into a list
	inline static void call(Func *functor, ...){
		//Vector used to collect arguments
		std::vector<int> templateParamList;

		//Variable argument list manip.
		va_list listPointer;
		va_start( listPointer, functor );
		//Collect parameters into the list
		for( int i = 0 ; i < NumArguments; i++ ){
			int val=va_arg( listPointer, int );
			templateParamList.push_back( val );
		}
		va_end( listPointer );

		//Call the actual typelist building function
		call( functor, templateParamList );
	}

	//Actual function called recursively to build a typelist based
	//on a list of values
	inline static void call( Func *functor, 
		std::vector<int> &templateParamList){

#			pragma message("KernelCaller::call") 

			//Get current parameter value in the list
			int val=templateParamList[templateParamList.size()-1];
			templateParamList.pop_back();
			
#if USE_BOOLEAN_SELECTION==0
			
			//Select the compile time value to add into the typelist
			//depending on the runtime variable and make recursive call. 
			//Recursive value selection mechanism.
			//+ make the next recursive call here
			ValueSelector<TList, NumArguments, Func, MaxArgValue, MaxArgValue>
				::select(functor, templateParamList, val);

#else	//Boolean selection
			
			//Select the compile time value to add into the typelist
			//depending on the runtime variable and make recursive call. 
			//Both versions are really instanciated.
			if(val){
				KernelCaller< 
				   Loki::Typelist<typename Loki::Int2Type<true>, TList >, 
				   NumArguments-1, Func, MaxArgValue >
				::call(functor, templateParamList);
			}else{
				KernelCaller< 
				   Loki::Typelist<typename Loki::Int2Type<false>, TList >, 
				   NumArguments-1, Func, MaxArgValue >
				::call(functor, templateParamList);
			}

#endif
	}
};

//Specialization for 0 value left in the list
//-> actual kernel functor call
template< class TList, class Func, int MaxArgValue>
struct KernelCaller<TList, 0, Func, MaxArgValue> {

	inline static void call( Func *functor ){
		//Call to the functor's kernel call method
		functor->call(TList());	//TList instanciated to get the method template parameter resolved.
	}
	inline static void call( Func *functor, 
		std::vector<int> &templateParams){

#	pragma message("KernelCaller::call FINAL") 
			functor->call (TList()); 
	}
};


//Value selector implementation
template<typename TList, int NumArguments, class Func, int MaxArgValue, int CurVal>
inline void ValueSelector<TList, NumArguments, Func, MaxArgValue, CurVal>::select(Func *functor, std::vector<int> &templateParamList, int val) {

#	pragma message("ValueSelector::select") 

	if(val==CurVal){
		KernelCaller<Loki::Typelist<typename Loki::Int2Type<CurVal>, TList >, NumArguments-1, Func, MaxArgValue >::call(functor, templateParamList);
	}else{
		ValueSelector<TList, NumArguments, Func, MaxArgValue, CurVal-1>::select(functor, templateParamList, val);
	}
}

#endif